#ifndef PEX8311_H
#define PEX8311_H 1

/******************************************************************/
/*                                                                */
/*                    The PLX PEX8311 header file                 */
/*                                                                */
/******************************************************************/
/*                                                                */
/* File Name   : PEX8311.h                                        */
/* Title       : The PEX 8311 header file                         */
/* Description : Defines the registers of the PEX8311 chip        */
/* Authors     : Alexey Shutov, Ilja Slepnev                      */
/* Revisions   :                                                  */
/*                                                                */
/******************************************************************/

/******************************************************************/
/*                                                                */
/*               Registers' structure definition                  */
/*                                                                */
/******************************************************************/

#include <linux/types.h>

typedef struct {
  unsigned int    unused    : 30;
  unsigned char   loc_reset : 1;
  unsigned int    unused1   : 1;
} mcbits;     //bridge control bits

typedef union {
  mcbits       bit;
  uint32_t     all;
} manycontrols;

//typedef struct {
//  unsigned char  g0dat     : 1;
//  unsigned char  g1dat     : 1;
//  unsigned char  g2dat     : 1;
//  unsigned char  g3dat     : 1;
//  unsigned char  g0oe      : 1;  //oe - output enable
//  unsigned char  g1oe      : 1;
//  unsigned char  g2oe      : 1;
//  unsigned char  g3oe      : 1;

//  unsigned char  g0ie      : 1;  //ie - interrupt enable
//  unsigned char  g1ie      : 1;
//  unsigned char  g2ie      : 1;
//  unsigned char  g3ie      : 1;

//  unsigned char  diag      : 2;
//  unsigned int  reserved  : 18;
//} gpiobits;

//typedef union {
//  gpiobits       bit;
//  unsigned       all;
//} gpio;

//typedef struct {
//  uint32_t     unused2[17];
//  uint32_t     pmcs;
//  uint8_t      unused4[10];
//  uint16_t     msic;
////  uint8_t      unused[0x1000-72];
//  uint8_t      unused[0x1000-84];
////end of 0x1000 space
//  uint32_t     unused1[6];    //dev init,interupt etc.
//  uint32_t     irs;           //interrupt reqest status
//  uint32_t     power;         //powerstatus
//  gpio         gpio;
//  uint32_t     unused3[3];
//  uint32_t     mb[4];
//  uint8_t      empty[0x1000-64];
////end of 0x2000 space
//} pex8311_peregs;           //pci-express registers


typedef struct {
  unsigned char  datw      : 2;  //data width def=0x3 32-bits
  unsigned char  wscount   : 4;  //wait state counter 0-15
  unsigned char  rie       : 1;  //ready input enable
  unsigned char  contbe    : 1;  //continuous burst enable
  unsigned char  lbe       : 1;  //local burst enable
  unsigned char  sgmode    : 1;  //scatter/gather mode enable
  unsigned char  doneie    : 1;  //done interrupt enable
  unsigned char  locam     : 1;  //local address mode 1-fifo(does not increment addr.)

  unsigned char  demm      : 1;  // demand mode enable
  unsigned char  waim      : 1;  //write and invalidate mode
  unsigned char  eoten     : 1;  //eot# ball enable
  unsigned char  fstm      : 1;  //fast/slow terminate mode

  unsigned char  ccm       : 1;  // clear count mode (for scatt./gath mode)
  unsigned char  intsel    : 1;  // 1-pcie interrupt, 0-local int.
  unsigned char  dcl       : 1;  // dac chain mode ? ? ? ?
  unsigned char  eotel     : 1;  // eot# end link(scat./gath. mode only)
  unsigned char  rmvme     : 1;  // ring management valid mode enable ?
  unsigned char  rmvsc     : 1;  // ring management valid stop control ?

  unsigned int  reserved  : 10;
} dmamodebits;

typedef union {
  dmamodebits       bit;
  unsigned int      all;
} dmamode;

typedef struct {
  unsigned char  dmace     : 1;  //chanel enable
  unsigned char  dmacs     : 1;  //chanel start
  unsigned char  dmaca     : 1;  //chanel abort
  unsigned char  dmaci     : 1;  //write 1 to clear interrupt
  unsigned char  dmadone   : 1;  //reading 1 - dma finished
  unsigned char  unused    : 3;  //
} dmacsbits;

typedef union {
  dmacsbits         bit;
  unsigned char     all;
} dmacs;

typedef struct {
  unsigned char  eis0       : 1;  //1 - enables LSERR# under detection of local parity error or pci abort
  unsigned char  eis1       : 1;  //not used
  unsigned char  genserrint : 1;  //writing 1 after 0 asserts SERR#
  unsigned char  mbie       : 1;  //mailbox int. enable
  unsigned char  pmie       : 1;  //power managment int. enable(LINTo#)
  unsigned char  pmi        : 1;  //poer management interrupt (write 1 to clear)
  unsigned char  dmwrpce    : 1;  //direct master w/r loc. data prity check error enable
  unsigned char  dmwrpces   : 1;  //direct master w/r loc. data prity check error status(1 to clear)

  unsigned char  intae      : 1;  //write 1 to enable INTA#(default 1)
  unsigned char  pcidie     : 1;  //enable loc->pci doorbell int.
  unsigned char  pciaie     : 1;  //pci abort int. enable
  unsigned char  liie       : 1;  //local int. input enable(LINTi# -> INTA#)
  unsigned char  rae        : 1;  //retry abort enable
  unsigned char  pcidia     : 1;  // pci doorbell int. active(ro)
  unsigned char  pciaia     : 1;  // pci master or target abort abort int. active(ro)
  unsigned char  lia        : 1;  // local int. active(ro)

  unsigned char  lioe       : 1;  // local int. output enable(default 1)
  unsigned char  ldie       : 1;  // pci->loc doorbell int. enable ?
  unsigned char  dma0ie     : 1;  // dma chanel 0 int. enable
  unsigned char  dma1ie     : 1;  // dma chanel 1 int. enable
  unsigned char  ldia       : 1;  // loc. doorbell int. active(ro)
  unsigned char  dma0ia     : 1;  // dma chanel 0 int. active(ro)
  unsigned char  dma1ia     : 1;  // dma chanel 1 int. active(ro)
  unsigned char  bistia     : 1;  // built-in self test int. active

  unsigned char  mtab       : 1;  // reading 1 inticates dir.mast. was bus master during m/t abort
  unsigned char  dma0ab     : 1;  // reading 1 inticates dma0 was bus master during m/t abort
  unsigned char  dma1ab     : 1;  // reading 1 inticates dma1 was bus master during m/t abort
  unsigned char  pexab      : 1;  // reading 1 inticates pex asserted  target abort after 256 master retry
  unsigned char  mbdat1     : 1;  // reading 1 indicates data in mbox1
  unsigned char  mbdat2     : 1;  // reading 1 indicates data in mbox2
  unsigned char  mbdat3     : 1;  // reading 1 indicates data in mbox3
  unsigned char  mbdat4     : 1;  // reading 1 indicates data in mbox4
} lintcsrbits;

typedef union {
  lintcsrbits           bit;
  unsigned int          all;
} lintcsr;

typedef struct {
  uint32_t     lcs_las0rr;
  uint32_t     lcs_las0ba;
  uint32_t     unused[14];
  uint32_t     mbox0;
  uint32_t     mbox1;
  uint32_t     mbox2;
  uint32_t     unused1[7];
  lintcsr      lintcsr;
  manycontrols manycontrols;
  uint32_t     unused2[4];
  dmamode      dmamode0;
  uint32_t     dmapcia0;       //dma pci address
  uint32_t     dmala0;         //dma local address
  uint32_t     dmats0;         //dma transfer size (bytes)
  uint32_t     dmadp0;         //dma descriptor pointer

  dmamode      dmamode1;
  uint32_t     dmapcia1;       //dma pci address
  uint32_t     dmala1;         //dma local address
  uint32_t     dmats1;         //dma transfer size (bytes)
  uint32_t     dmadp1;         //dma descriptor pointer

  dmacs        dmacs0;        //dma control/status
  dmacs        dmacs1;        //dma control/status
  uint8_t      unused3[2];
  uint32_t     dmaarb;         //dma arbitration


} pex8311_locregs;           //local registers

#endif /* PEX8311_H */
