
#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 33)
#include <linux/autoconf.h>
#else
#include <generated/autoconf.h>
#endif
#include <linux/module.h>
#include <linux/version.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/delay.h>
#include <linux/cdev.h>

#include "main.h"
#include "mlink.h"
#include "debug.h"
#include "mlink_op.h"
#include "mlink_op_pex.h"
#include "ml_proc.h"
#include "ml_fops.h"
#include "mlink_pci_ids.h"
#include "version.h"
#include "version_str.h"

MODULE_AUTHOR ("Alexey Shutov, Ilja Slepnev");
MODULE_LICENSE ("GPL");
MODULE_ALIAS ("pci:v000010B5d00009056sv000010EEsd0000A604bc*sc*i*");
MODULE_VERSION (VERSION_STR);

/* These identify the driver base version and may not be removed. */
static char version[] =
ALL_MSG ML_DEVICE_NAME " character device driver " VERSION_STR "\n";

int device_count = 0;    /* Number of cards detected */
struct mlink_dev_t dvtbl[DEVICE_MAX];
unsigned int device_major = 0; // Set to 0 to be dynamically assigned
static char class_name[] = ML_DEVICE_NAME;
static struct class *my_class;

static void setup_cdev(dev_p dev, int device_minor)
{
    int err, devno = MKDEV(device_major, device_minor);

    cdev_init(&dev->cdev, &dev_fops);
    dev->cdev.owner = THIS_MODULE;
    dev->cdev.ops = &dev_fops;
    err = cdev_add (&dev->cdev, devno, 1);
    /* Fail gracefully if need be */
    if (err)
        printk(KERN_NOTICE "Error %d adding %s%d", err, ML_DEVICE_NAME, device_minor);
}

/* ---------------------------------------------------------------------
 * --- Module initialization and cleanup (used at insmod/rmmod time) ---
 * ---------------------------------------------------------------------
 */

void cleanup_module (void)
{
    mlink_proc_exit();
//#ifdef CONFIG_PROC_FS
//    remove_proc_entry (ML_PROCFS_NAME, NULL);
//#endif
    while (device_count) {
        int device_minor = device_count-1;
        cdev_del(&dvtbl[device_minor].cdev);
        device_destroy(my_class, MKDEV(device_major, device_minor)); /* delete device node under /dev */
        dv_free (&dvtbl[device_minor]);  /* Unmap the card and turn off IRQ */
        device_count--;
    }

    class_destroy(my_class);
    unregister_chrdev_region(MKDEV (device_major, 0), DEVICE_MAX);
}

int init_module (void)
{
    struct pci_dev *pci_dev = NULL;
    int err;
    dev_t dev = MKDEV(device_major, 0);

    pr_info ("%s" , version);

    if (device_major)
        err = register_chrdev_region(dev, DEVICE_MAX, ML_DEVICE_NAME);
    else {
        err = alloc_chrdev_region(&dev, 0, DEVICE_MAX, ML_DEVICE_NAME);
        device_major = MAJOR(dev);
    }
    if (err < 0)
        return err;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6,4,0)) || \
    (defined(RHEL_RELEASE_CODE) && (RHEL_RELEASE_CODE >= 2308)) /* RHEL 9.4 */
    my_class = class_create(class_name);
#else
    my_class = class_create(THIS_MODULE, class_name);
#endif

    if(IS_ERR(my_class)) {
        printk(KERN_ERR ALL_MSG "Failed to create class\n");
        unregister_chrdev_region(MKDEV (device_major, 0), DEVICE_MAX);
        return -1;
    }
    mlink_proc_init();

    device_count = 0;
    while (1) {
        int device_minor = device_count;
        dev_p dv;
        pci_dev = pci_get_subsys(PCI_VENDOR_ID_PLX, PCI_DEVICE_ID_PLX_9056, PCI_VENDOR_ID_XILINX, PCI_DEVICE_ID_CUSTOM_PEXML, pci_dev);
        if (pci_dev == NULL)
            break;
        else
            printk(KERN_INFO "Found PEXML PCI-Express device [%d]\n", device_count+1);

        if (device_count >= DEVICE_MAX) {
            printk (KERN_WARNING ALL_MSG "you seem to have more than %d devices\n", DEVICE_MAX);   /* Recompile with higher DEVICE_MAX. */
            break;
        }

        dv = &dvtbl[device_count];
        memset (dv, 0, sizeof (*dv));
        setup_cdev(dv, device_minor);
        device_create(my_class, NULL, MKDEV(device_major, device_minor), NULL, ML_DEVICE_NAME "%d", device_minor);

        spin_lock_init (&dv->lock);
        mutex_init(&dv->dev_mutex);
        dv->err_status = "Initializing";
        dv->pci_dev_9056 = pci_dev;

        dv->id = device_count;
        sprintf (dv->name, ML_DEVICE_NAME "%d", dv->id);
        init_waitqueue_head (&dv->waitq);
        device_count++;
        dv_initialize (dv);         /* Ignore failure to initialize and let open() fail */
    }

    /*
   * Do not install the module when 0 devices are detected.
   */
    if (device_count == 0) {
        printk (KERN_INFO ALL_MSG "No devices found.\n");
        cleanup_module ();
        return -ENODEV;
    }
//#ifdef CONFIG_PROC_FS
//    create_proc_read_entry(ML_PROCFS_NAME, 0, NULL, drv_procfile_read, NULL);
//#endif

    return 0;                     /* Success */
}



void dv_setup_pci (dev_p dv)
{
    pci_set_master (dv->pci_dev_9056);  /* Make sure bus mastering is enabled */
#ifdef DEVICE_PCI_LATENCY
    {
        unsigned char c = DEVICE_PCI_LATENCY;

        pci_write_config_byte (dv->pcidev, PCI_LATENCY_TIMER, c);
        dv->latency = (int) c;
    }
#endif
}

/* ----------------------------------------------------------------------------
 * --- dv_free() un-does everything that dv_initialize() did, once per card ---
 * ----------------------------------------------------------------------------
 */

void dv_free (dev_p dv)
{
    size_t i;
    if (dv->initialized) {
        PDEBUG ("Freeing device and removing interrupt handler on irq %d\n", dv->irq);

        dv_int_disable (dv);        /* Make sure no more interrupts are generated */
        dv->err_status = "No longer initialized";

        free_irq (dv->irq, dv);
        dv->initialized = 0;
        dv->irq = 0;
    }

    dv_unmap_mreg(dv);

    // free RX buffers
    for (i=0; i<dv->rxbuf.capacity; i++) {
        dv_fifo_ent_t *rxent = &dv->rxbuf.ent[i];
        if (rxent->buf)
            dma_free_coherent (&dv->pci_dev_9056->dev, rxent->size, rxent->buf, rxent->bus_addr);
        rxent->size = 0;
        rxent->buf = NULL;
        rxent->bus_addr = 0;
    }
    dv->rxbuf.capacity = 0;

    // free TX buffers
    if (dv->txbuf.virt_addr)
        dma_free_coherent (&dv->pci_dev_9056->dev, dv->txbuf.size, dv->txbuf.virt_addr, dv->txbuf.bus_addr);
    dv->txbuf.size = 0;
    dv->txbuf.virt_addr = 0;
    dv->txbuf.bus_addr = 0;
}

/*
 * Since it is not possible to print error message from the real-time
 * interrupt handler, this function will check if any new errors need
 * to be reported during the regular Linux part of interrupt handling.
 */

void report_errors (dev_p dv)
{
    if (dv->dma_errors != dv->dma_report) {
        dv->dma_report = dv->dma_errors;
        printk (KERN_ERR "%s: %s\n", dv->name, dv->err_status);
    }
    if (dv->pci_error == 1 && !dv->save_bit_pci_error)
        //    printk("Dev is offline\n");
        dv->save_bit_pci_error = dv->pci_error;
}

void dv_clear_status(dev_p dv)
{
    // all bits are reset by hardware on LRESET# assert
    dv->hw_error_bits.all = 0;
    dv->lcf_data.all = 0;
    dv->save_lcf_data.all = 0;
}

/*
 * Check for PCI Bus abort and parity errors.
 *
 * This used to be called from the hard real-time part o the interrupt
 * handler, and so printk's were not allowed.  Now that it is in the
 * Linux interrupt handler instead, it is safe to call printk() again.
 * To allow errors to appear immediately on the console, I'm including
 * a call to dv_report_errors() in this routine for now.  Just calling
 * printk directly would be fine too.
 */
void dv_interrupt_check_pci_errors (dev_p dv)
{
    unsigned short pci_status = 0;
    int pci_error = 0;

    pci_read_config_word (dv->pci_dev_9056, PCI_STATUS, &pci_status);

    if (pci_status & PCI_STATUS_PARITY) {
        dv->dma_errors ++;
        pci_error++;
        dv->err_status = "PCI: Parity error";
        report_errors (dv);
    }
    if (pci_status & PCI_STATUS_DETECTED_PARITY) {
        dv->dma_errors ++;
        pci_error++;
        dv->err_status = "PCI: Detected parity error";
        report_errors (dv);
    }
    if (pci_status & PCI_STATUS_SIG_TARGET_ABORT) {
        dv->dma_errors ++;
        pci_error++;
        dv->err_status = "PCI: Signaled target abort";
        report_errors (dv);
    }
    if (pci_status & PCI_STATUS_REC_TARGET_ABORT) {
        dv->dma_errors ++;
        pci_error++;
        dv->err_status = "PCI: Received target abort";
        report_errors (dv);
    }
    if (pci_status & PCI_STATUS_REC_MASTER_ABORT) {
        dv->dma_errors ++;
        pci_error++;
        dv->err_status = "PCI: Received master abort";
        report_errors (dv);
    }
    if (pci_status & PCI_STATUS_SIG_SYSTEM_ERROR) {
        dv->dma_errors ++;
        pci_error++;
        dv->err_status = "PCI: Signaled system error";
        report_errors (dv);
    }

    if (pci_error) {
        printk (KERN_ERR "%s: pci error\n", dv->name);
        dv_xmit_disable (dv);
    }

    if (pci_error) {
        /*
     * Clear error flags, if any were detected (writing a 1 back
     * to the same bit clears it.)
     */
        pci_write_config_word (dv->pci_dev_9056, PCI_STATUS, pci_status);
        dv->pci_error = 1;
    }
}

/*
 * This is the standard, non-real-time interrupt handler.  If several
 * cards share the same IRQ, or share an IRQ with some other piece of
 * hardware, Linux will take care of calling this function, once for
 * each (*device).
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
irqreturn_t irq_handler (int irq, void *device, struct pt_regs *regs)
#else
irqreturn_t irq_handler (int irq, void *device)
#endif
{
    dev_p dv = (dev_p) device;
    int irqret = 0;

    spin_lock (&dv->lock);
    irqret = dv_interrupt_locked (dv);
    spin_unlock (&dv->lock);

    if (irqret == 1)
        dv_interrupt_bottom_half (dv);
    if (irqret)
        return IRQ_HANDLED;
    else
        return IRQ_NONE;
}


/* End of main.c */
