//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef MLINK_DRIVER_MAIN_H
#define MLINK_DRIVER_MAIN_H

#include "mlink_dev.h"
#include <linux/poll.h>
#include <linux/irqreturn.h>
#include <linux/interrupt.h>

enum { DEVICE_MAX = 4 };           /* Maximum number of cards supported */
extern int device_count;    /* Number of cards detected */
extern struct mlink_dev_t dvtbl[DEVICE_MAX];
extern unsigned int device_major;
extern struct file_operations dev_fops;
//static struct proc_dir_entry *proc_entry;

/*
 * Internal functions to initialize, reset, handle interrupts, etc.
 */
int dv_initialize (dev_p dv);
void dv_unmap_mreg (dev_p dv);
void dv_free (dev_p dv);
void dv_reset (dev_p dv);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
irqreturn_t irq_handler (int irq, void *device, struct pt_regs *regs);
#else
irqreturn_t irq_handler (int irq, void *device);
#endif
#ifndef IRQF_SHARED
#define IRQF_SHARED SA_SHIRQ
#endif

int dv_interrupt_locked (dev_p dv);
void dv_interrupt_bottom_half (dev_p dv);

int dev_open (struct inode *inode, struct file *file);
int dev_close (struct inode *inode, struct file *file);
unsigned int dev_poll (struct file *file, poll_table * wait);
long dev_ioctl (struct file *file, unsigned int cmd, unsigned long arg);
void dv_setup_pci (dev_p dv);
void dv_free (dev_p dv);
void report_errors (dev_p dv);
void dv_clear_status(dev_p dv);
void dv_interrupt_check_pci_errors (dev_p dv);

#endif // MLINK_DRIVER_MAIN_H
