//
//    Copyright 2010 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 33)
#include <linux/autoconf.h>
#else
#include <generated/autoconf.h>
#endif

#include <linux/version.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/poll.h>
#include <linux/delay.h>

#include "main.h"
#include "debug.h"
#include "mlink_op.h"
#include "mlink_op_pex.h"
#include "ml_proc.h"
#include "version.h"

#include "ml_fops.h"

/* ---------------------------------------------------------
 * --- File operations supported on the character device ---
 * ---------------------------------------------------------
 */


int dev_open (struct inode *inode, struct file *file)
{
//    int minor = iminor(file->f_path.dentry->d_inode) & 0x0f;
    int minor = iminor(inode) & 0x0f;
    dev_p dv;

    if (minor >= device_count)
        return -ENXIO;
    dv = &dvtbl[minor];
    if (dv->pid)
        return -EBUSY;              /* Only allow one pid to open */
    if (!dv->initialized)
        return -EIO;                /* Hardware I/O error */

    dv->pid = current->pid;
    return 0;
}

int dev_close (struct inode *inode, struct file *file)
{
//    int minor = iminor(file->f_path.dentry->d_inode) & 0x0f;
    int minor = iminor(inode) & 0x0f;
    dev_p dv;

    dvtbl[minor].pid = 0;

    if (minor >= device_count)
        return -ENXIO;
    dv = &dvtbl[minor];

    if (dv && dv->initialized) {
        dv_xmit_disable (dv);
    }
    return 0;
}

/*
 * The dev_poll routine gets used when a program uses select().
 * At the moment, only things which cause an interrupt can wake
 * up a select() call. (ex: In the case of the SDSU/Motorola chip,
 * for example, this means you can select() to find out if DMA
 * is complete and also if there is anything to read() in FIFO
 * mode.  With the AMCC chip, it does not generate an interrupt
 * for FIFO mode transfers, so unless a periodic check was built
 * into the driver, one should only use select() on sspci when
 * a DMA transfer is in progress.)
 */

unsigned int dev_poll (struct file *file, poll_table * wait)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 19, 0)
    unsigned int minor = MINOR (file->f_dentry->d_inode->i_rdev);
#else
    unsigned int minor = MINOR (file->f_path.dentry->d_inode->i_rdev);
#endif
    dev_p dv = &dvtbl[minor];
    unsigned int mask = 0;
    int polled = 0;

    while (1) {
        //    printk("poll_wait() entered\n");
        if (dv->rtc_event) {
            mask |= (POLLWRNORM | POLLOUT);  // Bus-master read = write function
            dv->rtc_event = 0;
            //      printk("poll: rtc_event, mask 0x%08X\n", mask);
            //   break;
        }

        if (dv->wtc_event) {
            mask |= (POLLRDNORM | POLLIN);   // Bus-master Write = read function
            dv->wtc_event = 0;
            //      printk("poll: wtc_event, mask 0x%08X\n", mask);
            //   break;
        }

        // if (mask) printk("poll: mask %d %s\n", mask, polled?"(wait complete)":"");
        if (mask || polled)
            break;
        poll_wait (file, &dv->waitq, wait);
        polled = 1;
    }
    return mask;
}

struct file_operations dev_fops = {
    poll:dev_poll,
    unlocked_ioctl:dev_ioctl,
    open:dev_open,
    release:dev_close,
};
