#ifndef MLINK_FRAME_HEADER_H
#define MLINK_FRAME_HEADER_H

#ifdef __KERNEL__
#include <linux/types.h>
#else
#include <stdint.h>
#endif

// M-Link frame header
#pragma pack(push, 1)
typedef struct frame_hdr_t {
    uint16_t type; // Frame type. Transferred second.
    uint16_t sync; // Frame sync. Must be set equal to ML_FRAME_SYNC. Transferred first on 16-bit bus.
    uint16_t seq; // Frame sequence number. Incremented on every next frame sent in one direction.
    uint16_t len; // Frame length in 32-bit words. Valid length is 4 – 1023 (data size 0-1019)
    uint16_t src; // Source address
    uint16_t dst; // Destination address
} frame_hdr_t;
#pragma pack(pop)

enum { ML_FRAME_SYNC = 0x2A50 };
enum { ML_FRAME_HEADER_WORDS = 3 };
enum { ML_FRAME_TRAILER_WORDS = 1 };

#define ML_MIN_RAW_FRAME_WORDS (ML_FRAME_HEADER_WORDS+ML_FRAME_TRAILER_WORDS)

// Maximum number of words in frame including header, crc. See M-Link specification.
enum { ML_MAX_RAW_FRAME_WORDS = 511 };

// Maximum number of payload data words.
#define ML_MAX_FRAME_DATA_WORDS (ML_MAX_RAW_FRAME_WORDS-(ML_FRAME_HEADER_WORDS+ML_FRAME_TRAILER_WORDS))

#endif // MLINK_FRAME_HEADER_H
