#ifndef MLINK_OP_PEX_H
#define MLINK_OP_PEX_H

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 33)
#include <linux/autoconf.h>
#else
//#include <generated/autoconf.h>
#endif
#include <linux/types.h>
#include <linux/module.h>
#include <linux/uaccess.h>
#include <linux/io.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/spinlock.h>

#include "debug.h"
#include "mlink_dev.h"
#include "mlink_op.h"
#include "mlink_op_pex.h"
#include "pexml_ports.h"
#include "compat.h"

#ifndef MLINK_PEX
# define MLINK_PEX
#endif

void dv_write_disable (dev_p dv) /* RX */
{
    dv->mreg->dmacs1.bit.dmace = 0;      /* dma1 chanel disable */

    dv->mreg->dmacs1.bit.dmaca = 1;      /* dma1 chanel abort */
    dv->mreg->dmamode1.bit.doneie = 0;   /* Write interrupt disabled */

    dv->mreg->dmacs1.bit.dmacs = 0;      /* Disable bus mastering */
    dv->mreg->dmapcia1 = 0;              /* No address for DMA transfer */
    dv->mreg->dmats1 = 0;                /* Write count must be zero */
}

void dv_read_disable (dev_p dv) /* TX */
{
    dv->mreg->dmacs0.bit.dmace = 0;      /* dma0 chanel disable */
    dv->mreg->dmacs0.bit.dmaca = 1;      /* dma0 chanel abort */
    dv->mreg->dmamode0.bit.doneie = 0;   /* Read interrupt disabled */
    dv->mreg->dmacs0.bit.dmacs = 0;      /* Disable bus mastering */
    dv->mreg->dmapcia0 = 0;              /* No address for DMA transfer */
    dv->mreg->dmats0 = 0;                /* Read count must be zero */
}

void dv_xmit_disable (dev_p dv)
{
    dv_write_disable (dv);
    dv_read_disable (dv);
}

/// clear interrupt request on local bus
/// imb interrupt is disabled until dv_imb_irq_ack()
void ml_clear_imb_interrupt_request(dev_p dv)
{
    dv_set_irq_mask_bits(dv, 0, LOCAL_IRQ_SRC_MASK_RX_DMA_BIT | LOCAL_IRQ_SRC_MASK_LCF_BIT, OPTIMAL_UPDATE);
}

void dv_int_disable (dev_p dv)
{
    dv->mreg->dmacs0.bit.dmacs = 0;      /* Disable bus mastering */
    dv->mreg->dmacs1.bit.dmacs = 0;      /* Disable bus mastering */
    dv->mreg->dmapcia0 = 0;
    dv->mreg->dmats0 = 0;
    dv->mreg->dmapcia1 = 0;
    dv->mreg->dmats1 = 0;
    dv->mreg->lintcsr.bit.dma0ie=0;   //disabling interrupt on dma 0 complete
    dv->mreg->lintcsr.bit.dma1ie=0;   //disabling interrupt on dma 1 complete
    dv->mreg->lintcsr.bit.liie=0;     //disabling local interrupt for incoming mailbox interrupt

    ml_clear_imb_interrupt_request(dv);
}

// r/w hw regs in pexml board
void dv_write_pexml_hw_reg(dev_p dv, uint32_t data, uint32_t hw_addr)
{
    uint32_t *addr = dv->mem2 + (hw_addr & (HW_ADDR_AREA_SIZE/4-1));
    if (hw_addr >= HW_ADDR_AREA_SIZE/4) {
        PDEBUG("Error: Write HW REG %08X out of address range", hw_addr);
        return;
    }
    iowrite32(data, addr);
}

uint32_t dv_read_pexml_hw_reg(dev_p dv, uint32_t hw_addr)
{
    uint32_t *addr = dv->mem2 + (hw_addr & (HW_ADDR_AREA_SIZE/4-1));
    if (hw_addr >= HW_ADDR_AREA_SIZE/4) {
        PDEBUG("Error: Read HW REG %08X out of address range", hw_addr);
        return 0;
    }
    return ioread32(addr);
}

void dv_set_irq_mask_bits(dev_p dv, uint32_t bits_set, uint32_t bits_clear, int force_update)
{
    uint32_t old_bits = dv->irq_mask;
    uint32_t new_bits = old_bits;
    new_bits |= bits_set;
    new_bits &= ~bits_clear;
    if (force_update || (old_bits != new_bits))
        dv_write_pexml_hw_reg(dv, new_bits, HW_CSR_IRQ_SRC_EN_MASK_REG);
}

uint32_t dv_calc_rx_dma_w32_size(dev_p dv)
{
    uint32_t rx_wr_ptrram_ptr = dv_read_pexml_hw_reg(dv, HW_RX_WR_PTRRAM_PTR_REG); //read number of writen frames

    if (dv->rx_rd_ptrram_ptr == rx_wr_ptrram_ptr) {
        PDEBUG("Error: no packets available, rx_rd_ptr = rx_wr_ptr %u\n", rx_wr_ptrram_ptr);
        return 0;  //smth. wrong, interrupt impossible for such state
    } else {
        uint32_t size_w = 0;
        uint32_t first_ptrram_data = dv_read_pexml_hw_reg(dv, HW_RX_PTR_RAM_BASE_ADDR + dv->rx_rd_ptrram_ptr); // read first frame pointer
        uint32_t start_addr = (first_ptrram_data >> 16) & 0xFFFF;
        // unsafe optimization:  uint32_t start_addr = dv->rx_rd_dataram_ptr;
        if (dv->rx_rd_dataram_ptr != start_addr) {
            printk(KERN_ERR "%s: rx_rd_dataram_ptr %u != start_addr %u\n",
                   dv->name, dv->rx_rd_dataram_ptr, start_addr);
        }
        uint32_t last_ptr = (rx_wr_ptrram_ptr - 1) % dv->rx_ptrram_size;
        uint32_t last_ptrram_data  = dv_read_pexml_hw_reg(dv, HW_RX_PTR_RAM_BASE_ADDR + last_ptr); // read last frame pointer
        uint32_t stop_addr = last_ptrram_data & 0xFFFF;
        if (start_addr > dv->rx_dataram_size) {
            printk(KERN_ERR "%s: start_addr %u > dv->rx_dataram_size %u",
                   dv->name, start_addr, dv->rx_dataram_size);
            dv->stats.rx_dropped++;
            return 0;
        }
        if (stop_addr > dv->rx_dataram_size) {
            printk(KERN_ERR "%s: stop_addr %u > dv->rx_dataram_size %u",
                   dv->name, stop_addr, dv->rx_dataram_size);
            dv->stats.rx_dropped++;
            return 0;
        }
        size_w = (start_addr < stop_addr) ? (stop_addr - start_addr + 1) : (dv->rx_dataram_size + stop_addr - start_addr + 1);
        // statistics
        {
            // circular buffer size = (capacity + head - tail) % capacity
            uint32_t rx_packets = (dv->rx_ptrram_size + rx_wr_ptrram_ptr - dv->rx_rd_ptrram_ptr) % dv->rx_ptrram_size;
            uint32_t rx_bytes = size_w * 4;
            PDEBUGG("%s: HW RX: packet #%8lu, ptrram %3u - %3u: %3u frames, %5u bytes -> rxbuf[%2u]",
                   dv->name, dv->stats.rx_packets,
                   (dv->rx_rd_ptrram_ptr+1) % dv->rx_ptrram_size, rx_wr_ptrram_ptr, rx_packets, rx_bytes, dv->rxbuf.ndirty);
            dv->stats.rx_packets += rx_packets;
            dv->stats.rx_bytes += rx_bytes;
        }
        // advance read pointer
        dv->rx_rd_ptrram_ptr = rx_wr_ptrram_ptr;
        return size_w;
    }
}

void dv_recv_imb (dev_p dv)
{
    dv->imb1.all = dv_read_pexml_hw_reg(dv, HW_CSR_IRQ_SRC_REG);
}

int dv_send_dma (dev_p dv, ml_dmaop_ro_t dmaop)
{
    unsigned long flags;
    int err = 0;
    long wait_ret;
    int timeout_ticks;

    err = pexml_access_ok (VERIFY_READ, (const void *) dmaop.buf, dmaop.size);
    if (err < 0) {
        return (err);
    }
    if (dmaop.size <= 0) {
        printk (KERN_ERR "%s: bad transfer size: %ld\n", dv->name, dmaop.size);
        return (-EINVAL);
    }
    if (dmaop.size > dv->max_tx_frame_bytes) {
        printk (KERN_ERR "%s: requested transfer size too large: %ld, maximum is %ld\n", dv->name, dmaop.size, dv->max_tx_frame_bytes);
        return (-ENOMEM);
    }
    if (!txbuf_has_space(dv, dmaop.size)) {
        printk (KERN_ERR "%s: Tx buffer full: no space for %ld bytes\n", dv->name, dmaop.size);
        return (-ENOMEM);
    }

    // check if txbuf not busy and get it
    {
        int txbuf_not_available = 0;
        spin_lock_irqsave (&dv->lock, flags);
        if (dv->txbuf.flags & DMA_BUF_FLAG_BUSY)
            txbuf_not_available = 1;
        else
            dv->txbuf.flags |= DMA_BUF_FLAG_BUSY;
        spin_unlock_irqrestore (&dv->lock, flags);
        if (txbuf_not_available) {
            printk (KERN_ERR "%s: TX buffer busy\n", dv->name);
            return (-EBUSY);
        }
    }

    // check for previous frame was send successfuly in hardware (tx_wr_pointers == tx_rd_pointers)
    // can check only ptrram pointer, as they both(dataram and prtram) write simultaneously
    {
        uint32_t tx_rd_ptrram_ptr = dv_read_pexml_hw_reg(dv, HW_TX_RD_PTRRAM_PTR_REG);
        if (dv->tx_wr_ptrram_ptr != tx_rd_ptrram_ptr) {
            printk (KERN_ERR "%s: TX HW busy, tx_wr_ptrram=%08X, tx_rd_ptrram=%08X\n", dv->name, dv->tx_wr_ptrram_ptr, tx_rd_ptrram_ptr);
            return (-EBUSY);
        }
    }
    //      printk (KERN_ERR "DMA Tx: %08x @ %08x\n", dmaop.size, dv->txbuf.bus_addr);

    if (copy_from_user (dv->txbuf.virt_addr, dmaop.buf, dmaop.size))
        return (-EFAULT);

    spin_lock_irqsave (&dv->lock, flags);
    dv->txbuf.flags &= ~DMA_BUF_FLAG_COMPLETE;
    dv->txbuf.count = 0;
    dv->txbuf.used = dmaop.size;
    dv->txbuf.seq++;
    dv->mreg->dmamode0.bit.doneie = 0;   //disable int. on dma done
    dv->mreg->dmapcia0 = dv->txbuf.bus_addr;
    dv->mreg->dmala0 = (((HW_TX_DATA_RAM_BASE_ADDR | dv->tx_wr_dataram_ptr) << 2)
                        & (HW_ADDR_AREA_SIZE-1));   //for DMA need conv. to addr[31:0], not [31:2]
    dv->mreg->dmats0 = dmaop.size;
    dv->mreg->dmamode0.bit.doneie = 1;   //int. on dma done
    dv->mreg->dmacs0.bit.dmace = 1;
    dv->mreg->dmacs0.bit.dmacs = 1;
    spin_unlock_irqrestore (&dv->lock, flags);

    if (dmaop.timeout == -1)
        return 0;

    timeout_ticks = (dmaop.timeout*HZ)/1000;
    // event_condition: 1 to end wait loop
#define tx_event_condition (dv->txbuf.flags & DMA_BUF_FLAG_COMPLETE)
    if (timeout_ticks == 0)
        wait_ret = wait_event_interruptible (dv->waitq, tx_event_condition);
    else
        wait_ret = wait_event_interruptible_timeout (dv->waitq, tx_event_condition, timeout_ticks);

    if (wait_ret == 0) {
        u32 irq_src, irq_mask;
        spin_lock_irqsave (&dv->lock, flags);
        irq_src = dv_read_pexml_hw_reg(dv, HW_CSR_IRQ_SRC_REG);
        irq_mask = dv_read_pexml_hw_reg(dv, HW_CSR_IRQ_SRC_EN_MASK_REG);
        spin_unlock_irqrestore (&dv->lock, flags);
        printk (KERN_ERR "%s: DMA TX timeout, irq_src=%08X, irq_mask=%08X\n", dv->name, irq_src, irq_mask);
        dv->stats.tx_errors++;
        spin_lock_irqsave (&dv->lock, flags);
        txbuf_clear(dv);
        spin_unlock_irqrestore (&dv->lock, flags);
        return (-ETIME);
    }
    if (wait_ret == -ERESTARTSYS) {
        printk (KERN_NOTICE "%s: Transmit interrupted by signal\n", dv->name);
        return (-ERESTARTSYS);
    }
    //      if (signal_pending (current)) return -ERESTARTSYS;
    spin_lock_irqsave (&dv->lock, flags);
    {
        uint32_t ptrram_data = 0;
        uint32_t prev_tx_wr_dataram_ptr = 0;
        prev_tx_wr_dataram_ptr = dv->tx_wr_dataram_ptr;    // save start addr. for ptr_ram data
        dv->tx_wr_dataram_ptr = (dv->tx_wr_dataram_ptr + dmaop.size/4) % dv->tx_dataram_size; // calc. new dataram wr addr(conv. bytes to 32bit words)
        dv_write_pexml_hw_reg(dv, dv->tx_wr_dataram_ptr, HW_TX_WR_DATARAM_PTR_REG); //save data ptr. in dataram ptr. reg
        ptrram_data = ((dv->tx_wr_dataram_ptr - 1)&0xFFFF) | ((prev_tx_wr_dataram_ptr&0xFFFF)<<16); //data for ptrram
        dv_write_pexml_hw_reg(dv, ptrram_data, HW_TX_PTR_RAM_BASE_ADDR + dv->tx_wr_ptrram_ptr); //write ptrram
        dv->tx_wr_ptrram_ptr = (dv->tx_wr_ptrram_ptr + 1) % dv->tx_ptrram_size; //calc new ptrram addr
        dv_write_pexml_hw_reg(dv, dv->tx_wr_ptrram_ptr, HW_TX_WR_PTRRAM_PTR_REG);   //write ptrram addr. reg
        txbuf_clear(dv);
    }
    spin_unlock_irqrestore (&dv->lock, flags);
    return 0;
}

int dv_poll_event(dev_p dv, ml_poll_ev_op_t *op)
{
    long wait_ret = 0;
    // event_condition: 1 to end wait loop
#define hws_event_condition ( \
    (dv->lcf_data.all != dv->save_lcf_data.all) \
            || (dv->rxbuf.nwr - dv->rxbuf.nrd > 0UL))
wait_ret = wait_event_interruptible_timeout (dv->waitq, hws_event_condition, op->timeout_ms);

    if (wait_ret == -ERESTARTSYS) {
        printk (KERN_NOTICE "%s: wait_event interrupted by signal\n", dv->name);
        return (-ERESTARTSYS);
    }
    op->mask = 0;
    {
        lcf_data_t lcf_data;
        lcf_data.all = dv->lcf_data.all;
        if (!dv->save_lcf_data.bit.spill && lcf_data.bit.spill) op->mask |= ML_EV_BIT_SPILL_ON;
        if (dv->save_lcf_data.bit.spill && !lcf_data.bit.spill) op->mask |= ML_EV_BIT_SPILL_OFF;
        rxfifo_ptr nent = dv->rxbuf.nwr - dv->rxbuf.nrd;
        if (nent > 0UL)
            op->mask |= ML_EV_BIT_RX_PKT;
        dv->save_lcf_data.all = lcf_data.all;
    }
    return 0;
}

// run from bottom-half interrupt handler
void dv_receive_packets(dev_p dv)
{
    unsigned long flags;
    // write-transfer-complete interrupt serialization
    // don't fire new DMA until we have served WTC interrupt
    if (dv->mreg->dmacs1.bit.dmace) {
        PDEBUG("%s: dv_receive_packets: dmace=1 (DMA WTC interrupt not served yet)", dv->name);
        return;
    }

    uint32_t rxw = dv_calc_rx_dma_w32_size(dv);
    uint32_t rxb = rxw * 4;
    if (rxw == 0) {
        PDEBUG("Unexpected rxcnt interrupt, rxw = 0\n");
        return;
    }
    if (rxw > dv->rx_dataram_size) {
        printk(KERN_ERR "%s: Bad rxw = %d > %d\n",
               dv->name, rxw, dv->rx_dataram_size);
        return;
    }

    if (dv->rx_request_size_bytes > 0) {
        printk(KERN_ERR "%s Unexpected rxcnt interrupt, previous is not served\n",
               dv->name);
        dv->stats.rx_over_errors++;
    }
    dv->rx_request_size_bytes += rxb; // FIXME: Issue#265


    PDEBUGG ("BH: trying start RX DMA for %d bytes\n", rxb);
    // range check and start RX DMA

    uint32_t dma_xfer_bytes = rxb;
    uint32_t rxbuf_index = dv->rxbuf.ndirty % dv->rxbuf.capacity;
    dv_fifo_ent_t *rxent = &dv->rxbuf.ent[rxbuf_index];
    const uint32_t allocated_size = rxent->size;
    if (dma_xfer_bytes > allocated_size) {
        printk(KERN_ERR "%s: RX frame too large: %u truncated to %u\n",
               dv->name, dma_xfer_bytes, allocated_size);
        dma_xfer_bytes = allocated_size;
        // FIXME: reset rx fifo
    }

    // run locked
    spin_lock_irqsave(&dv->lock, flags);
    const unsigned int dirty_avail = rxbuf_dirty_avail(dv);

    if (dirty_avail > 0) {
        dv->rxbuf.ndirty = (dv->rxbuf.ndirty + 1) % dv->rxbuf.capacity;

        rxent->len = dma_xfer_bytes;

        // FIXME: check if dma channel is not active
        dv->mreg->dmamode1.bit.doneie = 0;
        dv->mreg->dmacs1.bit.dmace = 1;
        dv->mreg->dmacs1.bit.dmacs = 0;

        dv->mreg->dmala1 = (((HW_RX_DATA_RAM_BASE_ADDR | dv->rx_rd_dataram_ptr) << 2)
                            & (HW_ADDR_AREA_SIZE-1));   //for DMA need conv. to addr[31:0], not [31:2]
        dv->mreg->dmapcia1 = rxent->bus_addr;
        dv->mreg->dmats1 = rxent->len;
        dv->mreg->dmamode1.bit.doneie = 1;
        dv->mreg->dmacs1.bit.dmacs = 1;   /* Enable bus mastering */

        dv->rx_rd_dataram_ptr = (dv->rx_rd_dataram_ptr + (dv->rx_request_size_bytes/4)) % dv->rx_dataram_size;
            //we will write dv->rx_wr_dataram_ptr to HW on rx dma complete (dv_interrupt_locked_handle_wtc)
        dv->rx_request_size_bytes = 0;
        PDEBUGG("fired DMA %5u bytes, ndirty=%2u", rxent->len, dv->rxbuf.ndirty);
    }
    spin_unlock_irqrestore(&dv->lock, flags);

    if (dirty_avail <= 0) {
        printk(KERN_NOTICE "%s: RX buffer busy, %d bytes receive postponed (RX overrun error)\n", dv->name, rxb);
        dv->stats.rx_over_errors++;
    }
}

void update_link_status(dev_p dv)
{
    u32 rx_status, tx_status;
    rx_status = dv_read_pexml_hw_reg(dv, HW_CSR_RX_STATUS);
    tx_status = dv_read_pexml_hw_reg(dv, HW_CSR_TX_STATUS);
    dv->hw_error_bits.bit.rx_link_down = !(rx_status & 0x0001);
    dv->hw_error_bits.bit.lcf_timeout = rx_status & 0x0080;
    dv->hw_error_bits.bit.loss_of_signal = !(rx_status & 0x0100);
    dv->hw_error_bits.bit.laser_fault = (tx_status & 0x0008);
    dv->lcf_data.all = 0xFF & dv_read_pexml_hw_reg(dv, HW_RX_LCF_DATA_REG); //read lcf data
}

void update_board_id(dev_p dv)
{
    {
        u32 ow_id_lo = dv_read_pexml_hw_reg(dv, HW_CSR_OW_ID_LO_REG);
        u32 ow_id_hi = dv_read_pexml_hw_reg(dv, HW_CSR_OW_ID_HI_REG);
        u64 ow_id = ((u64)ow_id_hi << 32) | ow_id_lo;
        dv->board_info.board_serial = (ow_id >> 8) & 0xFFFFFFFFFFFFull;
    }
    {
        u32 fw_ver = dv_read_pexml_hw_reg(dv, HW_CSR_FW_VER_REG);
        dv->board_info.fw_version.fw_ver_maj = (fw_ver >> 8) & 0xFFFF;
        dv->board_info.fw_version.fw_ver_min = fw_ver & 0xFF;
        dv->board_info.fw_version.fw_rev = dv_read_pexml_hw_reg(dv, HW_CSR_FW_REV_REG);
    }
}

void update_board_info(dev_p dv)
{
    update_board_id(dv);
    {
        u32 board_config = dv_read_pexml_hw_reg(dv, HW_CSR_BOARD_CONFIG_REG);
        if (board_config != 0) {
            u32 tx_ram_bytes = 4 * (1<<(board_config & 0x1F));
            u32 tx_ram_packets = 1<<((board_config>>5) & 0x1F);
            u32 rx_ram_bytes = 4 * (1<<((board_config>>10) & 0x1F));
            u32 rx_ram_packets = 1<<((board_config>>15) & 0x1F);
            dv->tx_dataram_size = tx_ram_bytes/4;
            dv->rx_dataram_size = rx_ram_bytes/4;
            dv->tx_ptrram_size = tx_ram_packets;
            dv->rx_ptrram_size = rx_ram_packets;
        } else {
            dv->tx_dataram_size = 1024;
            dv->rx_dataram_size = 8192;
            dv->tx_ptrram_size = 8;
            dv->rx_ptrram_size = 64;
        }
    }
}

void print_board_info(dev_p dv)
{
//    if (dv->board_serial != 0) {
        printk(KERN_INFO "%s: Board serial number: %04X-%04X-%04X\n", dv->name,
               (u16)((dv->board_info.board_serial >> 32) & 0xFFFF),
               (u16)((dv->board_info.board_serial >> 16) & 0xFFFF),
               (u16)(dv->board_info.board_serial & 0xFFFF));
//    }
    printk(KERN_INFO "%s: Firmware: ver %d.%d, rev %d\n", dv->name,
           dv->board_info.fw_version.fw_ver_maj,
           dv->board_info.fw_version.fw_ver_min,
           dv->board_info.fw_version.fw_rev);
    printk(KERN_INFO "%s: TX buffer: %d bytes, %d packets\n", dv->name,
           4 * dv->tx_dataram_size, dv->tx_ptrram_size);
    printk(KERN_INFO "%s: RX buffer: %d bytes, %d packets\n", dv->name,
           4 * dv->rx_dataram_size, dv->rx_ptrram_size);

}

int dv_write_port(dev_p dv, unsigned int port, u32 offset, u32 data)
{
    if (port < 1 || port > 4) {
        printk(KERN_ERR "%s: dv_write_port: invalid port %d\n", dv->name, port);
        return -EINVAL;
    }
    if (offset > 0x0F) {
        printk(KERN_ERR "%s: dv_write_port: invalid offset 0x%02X\n", dv->name, offset);
        return -EINVAL;
    }
    {
        u32 base = HW_CSR_REGION_ADDR + 0x10 * port;
        dv_write_pexml_hw_reg(dv, data, base + offset);
        if (0) printk(KERN_INFO "%s: dv_write_port: port %d, base %08x, offset %02X, data %08X %s%s%s%s%s%s\n",
               dv->name, port, base, offset, data,
               ((offset == PEXML_GTX_CTRL_REG) && (data & GTX_CTRL_RESET)) ? " GTX_RESET" : "",
               ((offset == PEXML_GTX_CTRL_REG) && (data & GTX_CTRL_LOOPBACK)) ? " GTX_LOOPBACK" : "",
               ((offset == PEXML_GTX_CTRL_REG) && (data & GTX_CTRL_RXPD)) ? " GTX_RXPD" : "",
               ((offset == PEXML_GTX_CTRL_REG) && (data & GTX_CTRL_TXPD)) ? " GTX_TXPD" : "",
               ((offset == PEXML_SFP_CTRL_REG) && (data & SFP_CTRL_RT_SEL)) ? " SFP_RT_SEL" : "",
               ((offset == PEXML_SFP_CTRL_REG) && (data & SFP_CTRL_TX_DISABLE)) ? " SFP_TX_DISABLE" : ""
               );
    }
    return 0;
}

int dv_read_port(dev_p dv, unsigned int port, u32 offset, u32 *data)
{
    if (port < 1 || port > 4) {
        printk(KERN_ERR "%s: dv_read_port: invalid port %d\n", dv->name, port);
        return -EINVAL;
    }
    if (offset > 0x0F) {
        printk(KERN_ERR "%s: dv_read_port: invalid offset 0x%02X\n", dv->name, offset);
        return -EINVAL;
    }
    {
        u32 base = HW_CSR_REGION_ADDR + 0x10 * port;
        u32 tmp = dv_read_pexml_hw_reg(dv, base + offset);
        // printk(KERN_INFO "%s: dv_read_port: port %d, offset %02X, data %08X\n", dv->name, port, offset, tmp);
        if (data)
            *data = tmp;
    }
    return 0;
}

int dv_caps_gtx_ctrl(dev_p dv)
{
    return (dv->board_info.fw_version.fw_rev == 1) || (dv->board_info.fw_version.fw_rev >= 25235);
}

int dv_flasher_present(dev_p dv)
{
    if (!dv) return 0;
    return (dv->board_info.fw_version.fw_rev >= 25239);
}

int dv_write_flasher(dev_p dv, u32 offset, u32 data)
{
    if (offset > HW_FLASHER_REGION_MASK) {
        printk(KERN_ERR "%s: %s: invalid offset 0x%02X\n", dv->name, __func__, offset);
        return -EINVAL;
    }
    if (!dv_flasher_present(dv)) {
        printk(KERN_ERR "%s: %s: flash programmer is not available on this firmware\n", dv->name, __func__);
        return -EINVAL;
    }
    dv_write_pexml_hw_reg(dv, data, HW_FLASHER_REGION_ADDR + offset);
    return 0;
}

int dv_read_flasher(dev_p dv, u32 offset, u32 *data)
{
    if (offset > HW_FLASHER_REGION_MASK) {
        printk(KERN_ERR "%s: %s: invalid offset 0x%02X\n", dv->name, __func__, offset);
        return -EINVAL;
    }
    if (!dv_flasher_present(dv)) {
        printk(KERN_ERR "%s: %s: flash programmer is not available on this firmware\n", dv->name, __func__);
        return -EINVAL;
    }
    u32 tmp = dv_read_pexml_hw_reg(dv, HW_FLASHER_REGION_ADDR + offset);
    if (data)
        *data = tmp;
    return 0;
}

#endif /* MLINK_OP_PEX_H */
